#include <windows.h>
#include <commctrl.h>
#include <math.h>
#include "..\ExDll\exdll.h"
#include "resource.h"
#include "LockedList.h"
#include "SystemEnum.h"

/* LockedList NSIS plug-in

  Title:   LockedList plug-in
  Author:  Afrow UK
  Version: 0.7
  Date:    26th February 2008

  Description:

   Plug-in that displays a list of open programs that are locking files that need to be
   modified/deleted etc by the installer.

  Changes:

  v0.7 - 26th February 2008
   * Re-wrote /autoclose code and fixed crashing.
   * Added AddClass and AddCaption functions.
   * Fixed Copy List memory read access error.
   * Made thread exiting faster for page leave.
   * Progress bar and % work better.
   * Processing mouse cursor redrawn.
   * Ignore button text only set when list is not empty.

  v0.6 - 12th February 2008
   * Added /autoclose "close_text" "kill_text" "failed_text" and
     /autoclosesilent "failed_text".
     The /ignore switch can be used along with this to set the Next
     button text.
   * Added /colheadings "application_text" "process_text"

  v0.5 - 25th November 2007
   * Fixed memory leak causing crash when re-visiting dialog. Caused
     by duplicate call to GlobalFree on the same pointer.

  v0.4 - 27th September 2007
   * Module or file names can now be just the file name as opposed to
     the full path.
   * Folder paths are converted to full paths (some are short DOS
     paths) before comparison.
   * Fixed typo in AddModule function (ModulesCount>FilesCount).
     Thanks kalverson.
   * List view is now scrolled into view while items are added.
   * List changed to multiple columns.
   * Debug privileges were not being set under SilentSearch.
   * Added /ignore switch that prevents the Next button being
     disabled.
   * Added AddApplications to add all running applications to the
     list.
   * Added processing mouse cursor.
   * Added right-click context menu with Close and Copy List options.
   * Added progress bar.
   * Added default program icon for processes without an icon.
   * Added code to resize controls for different dialog sizes.

  v0.3 - 13th July 2007
   * Added LVS_EX_LABELTIP style to list view control for long item texts.
   * Width of list header changed from width-6 to
     width-GetSystemMetrics(SM_CXHSCROLL).
   * Added WM_SYSMENU existence check when obtaining window captions.
   * Files/modules lists memory is now freed when using SilentSearch.
   * Files and Modules lists count now reset after a search.
   * Added reference to Unload function to read-me.

  v0.2 - 12th July 2007
   * Added two new examples.
   * Fixed pointer error in FileList struct causing only first module/file added to be used.
   * Fixed caption repetition over multiple processes.
   * Fixed stack overflow in DlgProc. Special thanks, Roman Prysiazhniuk for locating the source.
   * Better percent complete indication.

  v0.1 - 10th July 2007
   * First version.

*/

// Closes an application.
static void CloseApplication(HWND hWnd)
{
  SetActiveWindow(hWnd);
  SendMessageTimeout(hWnd, WM_CLOSE, 0, 0, SMTO_ABORTIFHUNG | SMTO_BLOCK, 1000, 0);
}

// Kills an application.
static void KillApplication(DWORD dwProcessId)
{
  HANDLE hProcess = OpenProcess(PROCESS_TERMINATE | SYNCHRONIZE, FALSE, dwProcessId);
  if (hProcess != NULL)
  {
    TerminateProcess(hProcess, 0xFFFFFFFF);
    WaitForSingleObject(hProcess, 1000);
    CloseHandle(hProcess);
  }
}

// Enable or disable installer page buttons.
static void EnableButtons(BOOL enable)
{
  EnableWindow(g_hNext, enable);
  EnableWindow(g_hBack, enable);
  EnableWindow(g_hCancel, enable);
}

// Handles the parent dialog.
static BOOL CALLBACK ParentDlgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  BOOL bRes = CallWindowProc((WNDPROC)ParentDlgProcOld, hWnd, uMsg, wParam, lParam);
  if (uMsg == WM_NOTIFY_OUTER_NEXT && !bRes)
  {
    // While the user waits, better disable the buttons.
    EnableButtons(FALSE);

    // Auto close.
    if (wParam == 1 && g_iRunning > 0 && g_bAutoClose)
      g_bAutoClosing = TRUE;
    else
    {
      // Determine which button was pressed.
      if (wParam == -1)
        g_is_back = TRUE;
      else if (wParam == NOTIFY_BYE_BYE)
        g_is_cancel = TRUE;
      g_done = TRUE;

      // End threads ASAP.
      FinishEnumeratingNow();
      PostMessage(g_hDialog, WM_CLOSE, 0, 0);
    }
  }
  return bRes;
}

// List window procedure.
LRESULT CALLBACK ListWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  switch (uMsg)
  {
    case WM_SETCURSOR:
    {
      if (g_hProcessingCursor)
      {
        SetCursor(g_hProcessingCursor);
        return TRUE;
      }
    }
  }
  return CallWindowProc(ListWndProcOld, hWnd, uMsg, wParam, lParam);
}

// Dialog procedure.
static LRESULT CALLBACK DlgProc(HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  switch (uMsg)
  {
  case WM_CONTEXTMENU:
    {
      if (wParam != (WPARAM)g_hList)
        break;

      // Get the point of mouse click on the list.
      POINT pt;
      if (lParam == ((UINT)-1))
      {
        RECT r;
        GetWindowRect(g_hList, &r);
        pt.x = r.left;
        pt.y = r.top;
      }
      else
      {
        pt.x = GET_X_LPARAM(lParam);
        pt.y = GET_Y_LPARAM(lParam);
      }

      // Determine which item was clicked on.
      LVITEM lvi;
      LVHITTESTINFO lvhti;
      lvhti.flags = LVHT_ONITEM;
      lvhti.pt.x = pt.x;
      lvhti.pt.y = pt.y;
      ScreenToClient(g_hList, &lvhti.pt);
      if ((lvi.iItem = ListView_HitTest(g_hList, &lvhti)) == -1)
        break;

      // Get that item.
      lvi.mask = LVIF_PARAM;
      lvi.iSubItem = 0;
      ListView_GetItem(g_hList, &lvi);

      if (!lvi.lParam)
        break;

      HMENU hMenu = CreatePopupMenu();
      if (((ITEM_TAG*)lvi.lParam)->hWnd > 0)
        AppendMenu(hMenu, MF_STRING, 1, "Close");
      AppendMenu(hMenu, MF_STRING, 2, "Copy List");
      int sel = TrackPopupMenu(hMenu, TPM_NONOTIFY|TPM_RETURNCMD, pt.x, pt.y, 0, g_hList, 0);
      // Close selected application.
      if (sel == 1 && lvi.lParam && IsWindow(((ITEM_TAG*)lvi.lParam)->hWnd))
      {
        CloseApplication(((ITEM_TAG*)lvi.lParam)->hWnd);
      }
      // Copy list to clipboard.
      else if (sel == 2)
      {
        int count = ListView_GetItemCount(g_hList);
        TCHAR *pszText = (TCHAR*)LocalAlloc(LPTR, sizeof(TCHAR)*256);
        TCHAR *pszSubText = (TCHAR*)LocalAlloc(LPTR, sizeof(TCHAR)*128);
        TCHAR *pszProcess = (TCHAR*)LocalAlloc(LPTR, sizeof(TCHAR)*512);
        TCHAR *pszFormat = (TCHAR*)LocalAlloc(LPTR, sizeof(TCHAR)*64);
        TCHAR *pszList = (TCHAR*)LocalAlloc(LPTR, sizeof(TCHAR)*512*count);

        lvi.pszText = pszText;
        lvi.cchTextMax = sizeof(TCHAR)*256;
        lvi.mask = LVIF_TEXT;

        ZeroMemory(pszList, sizeof(TCHAR)*512*count);
        
        // Calculate string lengths for tabulation.
        lvi.iSubItem = 0;
        int captionLen = 0;
        for (lvi.iItem=0; lvi.iItem<count; lvi.iItem++)
        {
          int len = 0;
          ListView_GetItem(g_hList, &lvi);
          if ((len = lstrlen(lvi.pszText)) > captionLen)
            captionLen = len;
        }
        wsprintf(pszFormat, "%%-2d [%%-4d] - %%-%ds - %%s\r\n", captionLen);

        // Add the items to the string.
        lvi.mask = LVIF_TEXT | LVIF_PARAM;
        for (lvi.iItem=0; lvi.iItem<count; lvi.iItem++)
        {
          lvi.iSubItem = 0;
          lvi.cchTextMax = sizeof(TCHAR)*256;
          lvi.pszText = pszText;
          ListView_GetItem(g_hList, &lvi);
          
          lvi.iSubItem = 1;
          lvi.cchTextMax = sizeof(TCHAR)*128;
          lvi.pszText = pszSubText;
          ListView_GetItem(g_hList, &lvi);

          wsprintf(pszProcess, pszFormat, lvi.iItem+1, ((ITEM_TAG*)lvi.lParam)->dwProcessId, pszText, pszSubText);
          lstrcat(pszList, pszProcess);
        }

        // Write to the clipboard.
        OpenClipboard(g_hWndParent);
        EmptyClipboard();
        SetClipboardData(CF_TEXT, pszList);
        CloseClipboard();

        // Free the allocated memory.
        LocalFree(pszList);
        LocalFree(pszFormat);
        LocalFree(pszProcess);
        LocalFree(pszSubText);
        LocalFree(pszText);
      }
      DestroyMenu(hMenu);
    }
  break;
  case WM_CTLCOLORSTATIC:
  case WM_CTLCOLOREDIT:
  case WM_CTLCOLORDLG:
  case WM_CTLCOLORBTN:
  case WM_CTLCOLORLISTBOX:
    return SendMessage(g_hWndParent, uMsg, wParam, lParam);
  }
  return FALSE;
}

// Add item not already on the list.
void AddItem(TCHAR *pszText, TCHAR *pszSubText, int len, HICON hIcon, ITEM_TAG *tag)
{
  int count = ListView_GetItemCount(g_hList);
  char *pszTextTemp = (char*)GlobalAlloc(GPTR, len);
  BOOL bAdded = FALSE;

  LVITEM lvi;
  lvi.mask = LVIF_TEXT | LVIF_PARAM;
  lvi.cchTextMax = len;
  lvi.pszText = pszTextTemp;
  lvi.iSubItem = 0;

  // Loop through existing items to check if the item has already been added.
  for (int i=0; i<count; i++)
  {
    lvi.iItem = i;
    ListView_GetItem(g_hList, &lvi);

    if (lvi.lParam && ((ITEM_TAG*)lvi.lParam)->dwProcessId == tag->dwProcessId && lstrcmp(lvi.pszText, pszText) == 0)
    {
      bAdded = TRUE;
      break;
    }
  }

  // Item wasn't previously added. Add it.
  if (bAdded)
    return;

  lvi.mask |= LVIF_IMAGE;
  lvi.pszText = pszText;
  lvi.iItem = count-1;
  lvi.lParam = (LPARAM)tag;

  // Add the executable icon to the image list.
  if (hIcon)
  {
    HIMAGELIST hImageList = ListView_GetImageList(g_hList, LVSIL_SMALL);
    int iIndex;

    iIndex = ImageList_AddIcon(hImageList, hIcon);
    DestroyIcon(hIcon);

    ListView_SetImageList(g_hList, hImageList, LVSIL_SMALL);

    lvi.iImage = iIndex;
  }
  else
    lvi.iImage = 2;

  // Add the item.
  ListView_InsertItem(g_hList, &lvi);

  // Set its sub item texts.
  lvi.mask = LVIF_TEXT;
  lvi.pszText = pszSubText;
  lvi.iSubItem = 1;
  ListView_SetItem(g_hList, &lvi);

  // Scroll the item into view.
  ListView_EnsureVisible(g_hList, count-1, FALSE);
}

// Ensures the NSIS installer has debug privileges.
BOOL EnableDebugPriv(void)
{
	HANDLE hToken;
	LUID sedebugnameValue;
	TOKEN_PRIVILEGES tkp;

	// Enable the SeDebugPrivilege
	if (!OpenProcessToken(GetCurrentProcess(),
		TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken))
		return FALSE;

	if (!LookupPrivilegeValue(NULL, SE_DEBUG_NAME, &sedebugnameValue))
  {
		CloseHandle(hToken);
		return FALSE;
	}

	tkp.PrivilegeCount = 1;
	tkp.Privileges[0].Luid = sedebugnameValue;
	tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;

	if (!AdjustTokenPrivileges(hToken, FALSE, &tkp, sizeof(tkp), NULL, NULL))
		return FALSE;
		
	CloseHandle(hToken);
  return TRUE;
}

// Callback for enumerating lock files.
BOOL CALLBACK EnumLockedFilesProc(FILE_INFORMATION File, LPARAM lParam)
{
  FileList *Files = (FileList*)lParam;
  static TCHAR szItemText[128];
  static TCHAR szProcessName[128];
  int iNewPercentComplete = Files->LastPC + int(Files->OutOfPC / float(File.TotalFiles) * float(File.FileNumber));

  // Update the progress.
  if (g_iPercentComplete != iNewPercentComplete)
  {
    LVITEM lvi;

    wsprintf(szItemText, "%d%% %s", iNewPercentComplete, (g_pszParams[2] ? g_pszParams[2] : TEXT_SEARCHING));

    lvi.mask = LVIF_TEXT;
    lvi.iItem = ListView_GetItemCount(g_hList)-1;
    lvi.iSubItem = 0;
    lvi.pszText = szItemText;
    ListView_SetItem(g_hList, &lvi);

    SendMessage(g_hProgress, PBM_SETPOS, iNewPercentComplete, 0);

    g_iPercentComplete = iNewPercentComplete;
  }

  if (File.ProcessId == GetCurrentProcessId())
    return TRUE;

  for (int i=0; i<Files->ListCount; i++)
  {
    if (g_done)
      return FALSE;

    if (!Files->Applications)
    {
      GetFullPathName(File.FullPath, MAX_PATH, File.FullPath, NULL);
      if (!EndsWith(CharLower(File.FullPath), Files->List[i]))
        continue;
    }
    else if (Files->Applications == 2)
    {
      TCHAR *wndcaption = (TCHAR*)LocalAlloc(LPTR, sizeof(TCHAR)*256);
      lstrcpy(wndcaption, File.ProcessCaption);
      CharLower(wndcaption);
      if (!wildcmp(Files->List[i], wndcaption))
      {
        LocalFree(wndcaption);
        continue;
      }
      LocalFree(wndcaption);
    }
    else if (Files->Applications == 3 && IsWindow(File.ProcessHWND))
    {
      TCHAR *wndclass = (TCHAR*)LocalAlloc(LPTR, sizeof(TCHAR)*256);
      GetClassName(File.ProcessHWND, wndclass, sizeof(TCHAR)*256);
      CharLower(wndclass);
      if (!wildcmp(Files->List[i], wndclass))
      {
        LocalFree(wndclass);
        continue;
      }
      LocalFree(wndclass);
    }

    if (*File.ProcessFullPath)
    {
      GetFileFromPath(File.ProcessFullPath, MAX_PATH, szItemText, 128);
      if (lstrcmpi(szItemText, "explorer.exe") == 0)
        continue;
    }
    else if (!*File.ProcessCaption && !*File.ProcessFullPath)
      wsprintf(szItemText, "Process Id: %d", File.ProcessId);
    else
      szItemText[0] = NULL;

    // Add list box item.
    if (Files->OutWhere == OutListBox)
    {
      HICON hIcon = NULL;

      // Get the icon for this process.
      if (*File.ProcessFullPath)
        hIcon = ExtractIcon(g_hInstance, File.ProcessFullPath, 0);

      // Add the new item if it does not already exist.
      ITEM_TAG *tag = (ITEM_TAG*)GlobalAlloc(GPTR, sizeof(ITEM_TAG));
      tag->hWnd = File.ProcessHWND;
      tag->dwProcessId = File.ProcessId;
      AddItem(File.ProcessCaption, szItemText, sizeof(szItemText), hIcon, tag);
    }
    // Add stack items.
    else
    {
      BOOL bAdded = FALSE;

      // Make sure we haven't already added this process.
      for (int j=0; j<ProcessCount; j++)
        if (ProcessIdsList[j] == File.ProcessId)
        {
          bAdded = TRUE;
          break;
        }

      // Process is new so add it.
      if (!bAdded)
      {
        ProcessIdsList[ProcessCount] = File.ProcessId;

        ProcessCaptionsList[ProcessCount]  = (TCHAR*)GlobalAlloc(GPTR, sizeof(TCHAR)*1024);
        ProcessFullPathsList[ProcessCount] = (TCHAR*)GlobalAlloc(GPTR, sizeof(TCHAR)*1024);

        lstrcpy(ProcessCaptionsList[ProcessCount], File.ProcessCaption);
        lstrcpy(ProcessFullPathsList[ProcessCount], File.ProcessFullPath);

        ProcessCount++;
      }
    }
  }
  return TRUE;
}

// Free allocated memory.
void FreeGlobalAllocs()
{
  int i;

  // Free allocated lists memory.
  for (i=FilesCount-1; i>=0; i--)
    if (FilesList[i])
      GlobalFree(FilesList[i]);

  for (i=ModulesCount-1; i>=0; i--)
    if (ModulesList[i])
      GlobalFree(ModulesList[i]);

  for (i=ClassesCount-1; i>=0; i--)
    if (ClassesList[i])
      GlobalFree(ClassesList[i]);

  for (i=CaptionsCount-1; i>=0; i--)
    if (CaptionsList[i])
      GlobalFree(CaptionsList[i]);

  FilesCount = ModulesCount = ClassesCount = CaptionsCount = 0;
  
  // Free allocated parameters memory.
  for (i=0; i<PARAMS; i++)
  {
    if (g_pszParams[i])
    {
      GlobalFree(g_pszParams[i]);
      g_pszParams[i] = NULL;
    }
  }
}

// Push all programs onto the stack.
void PushStrings()
{
  pushstring(OUT_END);
  while (--ProcessCount >= 0)
  {
    TCHAR szProcessId[8];
    wsprintf(szProcessId, "%d", ProcessIdsList[ProcessCount]);

    pushstring(ProcessCaptionsList[ProcessCount]);
    pushstring(ProcessFullPathsList[ProcessCount]);
    pushstring(szProcessId);

    GlobalFree(ProcessFullPathsList[ProcessCount]);
    GlobalFree(ProcessCaptionsList[ProcessCount]);
  }
  pushstring(OUT_START);

  // Free allocated memory.
  FreeGlobalAllocs();
}

// Loop through all processes to see if they are running.
BOOL CALLBACK EnumProcessIdsProc(DWORD dwProcessId, LPARAM lParam)
{
  if (((ITEM_TAG*)lParam)->dwProcessId == dwProcessId)
    return FALSE;
  return TRUE;
}

// Execute the system enumerations.
void SystemEnum(OUT_WHERE Where)
{
  POINT pt;
  FileList Files;
  Files.OutWhere = Where;

  // Get processing cursor.
  g_hProcessingCursor = (HCURSOR)LoadImage(NULL, IDC_APPSTARTING, IMAGE_CURSOR, 0, 0, LR_DEFAULTSIZE | LR_SHARED);

  // Redraw cursor the easy way.
  GetCursorPos(&pt);
  SetCursorPos(pt.x-1, pt.y);

  int r = 0;
  if (FilesCount)
    r = 5;
  if (ModulesCount)
    r++;
  if (ClassesCount)
    r++;
  if (CaptionsCount)
    r++;
  if (g_bShowRunningApplications)
    r++;
  int pc = 100 / r;

  // List all running applications first.
  if (g_bShowRunningApplications)
  {
    Files.List         = NULL;
    Files.ListCount    = 1;
    Files.OutOfPC      = pc;
    Files.LastPC       = SendMessage(g_hProgress, PBM_GETPOS, 0, 0);
    Files.Applications = TRUE;
    EnumApplications(EnumLockedFilesProc, (LONG)&Files);
  }

  // List all running applications first.
  if (CaptionsCount)
  {
    Files.List         = CaptionsList;
    Files.ListCount    = CaptionsCount;
    Files.OutOfPC      = pc;
    Files.LastPC       = SendMessage(g_hProgress, PBM_GETPOS, 0, 0);
    Files.Applications = 2;
    EnumApplications(EnumLockedFilesProc, (LONG)&Files);
  }

  // List all running applications first.
  if (ClassesCount)
  {
    Files.List         = ClassesList;
    Files.ListCount    = ClassesCount;
    Files.OutOfPC      = pc;
    Files.LastPC       = SendMessage(g_hProgress, PBM_GETPOS, 0, 0);
    Files.Applications = 3;
    EnumApplications(EnumLockedFilesProc, (LONG)&Files);
  }

  // List modules.
  if (ModulesCount)
  {
    Files.List         = ModulesList;
    Files.ListCount    = ModulesCount;
    Files.OutOfPC      = pc;
    Files.LastPC       = SendMessage(g_hProgress, PBM_GETPOS, 0, 0);
    Files.Applications = FALSE;
    EnumSystemModules(EnumLockedFilesProc, (LONG)&Files);
  }

  // List files.
  if (FilesCount)
  {
    Files.List         = FilesList;
    Files.ListCount    = FilesCount;
    Files.OutOfPC      = pc*5;
    Files.LastPC       = SendMessage(g_hProgress, PBM_GETPOS, 0, 0);
    Files.Applications = FALSE;
    EnumSystemHandles(EnumLockedFilesProc, (LONG)&Files);
  }

  // Set the progress bar position to 0.
  SendMessage(g_hProgress, PBM_SETPOS, 100, 0);

  // Redraw cursor the easy way.
  g_hProcessingCursor = NULL;
  GetCursorPos(&pt);
  SetCursorPos(pt.x-1, pt.y);
}

// Loop through files and modules.
DWORD WINAPI FilesThreadSilent(LPVOID)
{
  SystemEnum(OutStack);
  return FALSE;
}

// Loop through files and modules.
DWORD WINAPI FilesThread(LPVOID)
{
  LVITEM lvi;

  SystemEnum(OutListBox);

  g_iRunning = ListView_GetItemCount(g_hList)-1;
  lvi.iSubItem = 0;

  // Set Next button text.
  if (g_iRunning == 0)
    EnableWindow(g_hNext, TRUE);
  else if (g_pszParams[5])
  {
    if (g_pszParams[5][0])
    {
      g_pszParams[11] = (TCHAR*)GlobalAlloc(GPTR, sizeof(TCHAR)*256);
      GetWindowText(g_hNext, g_pszParams[11], sizeof(TCHAR)*256);
      SetWindowText(g_hNext, g_pszParams[5]);
    }
    EnableWindow(g_hNext, TRUE);
  }

  // Remove "Searching" item from list.
  ListView_DeleteItem(g_hList, g_iRunning);

  // Not finished searching but the user wants to leave the page.
  if (g_done)
  {
    lvi.mask = LVIF_TEXT | LVIF_IMAGE;
    lvi.iItem = g_iRunning;
    lvi.iImage = 0;
    lvi.pszText = (g_pszParams[3] ? g_pszParams[3] : TEXT_ENDSEARCH);
    ListView_InsertItem(g_hList, &lvi);
    ListView_EnsureVisible(g_hList, g_iRunning, FALSE);
  }
  // The list is not empty, so monitor running programs.
  else if (g_iRunning > 0)
  {
    lvi.mask = LVIF_PARAM;

    while (g_iRunning)
    {
      // Close programs on list.
      if (g_bAutoClosing)
      {
        TCHAR *pszCaption = (TCHAR*)LocalAlloc(LPTR, sizeof(TCHAR)*256);
        GetWindowText(g_hWndParent, pszCaption, sizeof(TCHAR)*256);

        if (g_bAutoClosing == TRUE)
        {
          if (g_bAutoClose == TRUE && MessageBox(g_hWndParent, (g_pszParams[7] && g_pszParams[7][0] ? g_pszParams[7] : TEXT_MSGBOX_CLOSE), pszCaption, MB_ICONEXCLAMATION | MB_YESNO) == IDNO)
          {
            EnableButtons(TRUE);
            g_bAutoClosing = FALSE;
          }
          else
          {
            for (lvi.iItem=0; lvi.iItem<g_iRunning; lvi.iItem++)
            {
              ListView_GetItem(g_hList, &lvi);
              if (lvi.lParam && IsWindow(((ITEM_TAG*)lvi.lParam)->hWnd))
                CloseApplication(((ITEM_TAG*)lvi.lParam)->hWnd);
            }
            Sleep(1000);
            g_bAutoClosing = 2;
          }
        }
        else if (g_bAutoClosing == 2)
        {
          if (g_bAutoClose == TRUE && MessageBox(g_hWndParent, (g_pszParams[8] && g_pszParams[8][0] ? g_pszParams[8] : TEXT_MSGBOX_KILL), pszCaption, MB_ICONEXCLAMATION | MB_YESNO) == IDNO)
          {
            EnableButtons(TRUE);
            g_bAutoClosing = FALSE;
          }
          else
          {
            for (lvi.iItem=0; lvi.iItem<g_iRunning; lvi.iItem++)
            {
              ListView_GetItem(g_hList, &lvi);
              if (lvi.lParam && ((ITEM_TAG*)lvi.lParam)->dwProcessId > 0)
                KillApplication(((ITEM_TAG*)lvi.lParam)->dwProcessId);
            }
            g_bAutoClosing = 3;
          }
        }
        else
        {
          MessageBox(g_hWndParent, (g_pszParams[6] && g_pszParams[6][0] ? g_pszParams[6] : TEXT_MSGBOX_ACFAILED), pszCaption, MB_ICONSTOP | MB_OK);
          EnableButtons(TRUE);
          g_bAutoClosing = FALSE;
        }

        LocalFree(pszCaption);
      } // if (g_bAutoClosing

      // Loop through programs on list.
      for (lvi.iItem=0; lvi.iItem<g_iRunning; lvi.iItem++)
      {
        if (g_done)
          break;

        ListView_GetItem(g_hList, &lvi);
        if (lvi.lParam)
        {
          // Check if process is still running.
          if (EnumProcessIds(EnumProcessIdsProc, lvi.lParam))
          {
            if (ListView_DeleteItem(g_hList, lvi.iItem))
            {
              GlobalFree((HGLOBAL)lvi.lParam);
              lvi.lParam = NULL;
              g_iRunning--;
              lvi.iItem--;
            }
          } // if (EnumProcessIds

        } // if (lvi.lParam
      } // for (lvi.iItem=

      for (int i=0; i<100; i++)
        if (g_done || g_iRunning == 0)
          goto exit_while;
        else if (!g_bAutoClosing)
          Sleep(10);
        else
          break;

    } // while (g_iRunning

exit_while:

    // Process monitoring was under way, but the user wants to leave the page.
    if (g_iRunning > 0)
    {
      lvi.mask = LVIF_TEXT | LVIF_IMAGE;
      lvi.iItem = g_iRunning;
      lvi.iImage = 0;
      lvi.pszText = (g_pszParams[4] ? g_pszParams[4] : TEXT_ENDMONITOR);
      ListView_InsertItem(g_hList, &lvi);
    }
  } // if (g_iRunning > 0

  // No programs found in the search,
  // or all programs once on the list have been closed.
  if (!g_done)
  {
    lvi.mask = LVIF_TEXT | LVIF_IMAGE;
    lvi.iItem = g_iRunning;
    lvi.iImage = 0;
    lvi.pszText = (g_pszParams[1] ? g_pszParams[1] : TEXT_NOPROGRAMS);
    ListView_InsertItem(g_hList, &lvi);

    g_iRunning = 0;

    if (g_bAutoClosing)
    {
      g_done = TRUE;
      g_bAutoClose = FALSE;
      PostMessage(g_hWndParent, WM_NOTIFY_OUTER_NEXT, 1, 0);
    }
    else
    {
      if (g_pszParams[11])
        SetWindowText(g_hNext, g_pszParams[11]);
      EnableWindow(g_hNext, TRUE);
    }
  } // if (!g_done

  ListView_EnsureVisible(g_hList, g_iRunning, FALSE);

  return FALSE;
}

// Creates our dialog.
void LoadDialog(int string_size)
{
  // Zero all memory first.
  for (int i=0; i<PARAMS; i++)
    g_pszParams[i] = NULL;

  // Get plugin parameters.
  TCHAR *pszParam = (TCHAR*)GlobalAlloc(GPTR, sizeof(TCHAR)*string_size);
  while (popstring(pszParam) == 0)
  {
    // Get heading text.
    if (lstrcmpi(pszParam, TEXT("/heading")) == 0)
    {
      g_pszParams[0] = (TCHAR*)GlobalAlloc(GPTR, sizeof(TCHAR)*string_size);
      popstring(g_pszParams[0]);
    }
    // Get no programs item text.
    else if (lstrcmpi(pszParam, TEXT("/noprograms")) == 0)
    {
      g_pszParams[1] = (TCHAR*)GlobalAlloc(GPTR, sizeof(TCHAR)*string_size);
      popstring(g_pszParams[1]);
    }
    // Get searching item text.
    else if (lstrcmpi(pszParam, TEXT("/searching")) == 0)
    {
      g_pszParams[2] = (TCHAR*)GlobalAlloc(GPTR, sizeof(TCHAR)*string_size);
      popstring(g_pszParams[2]);
    }
    // Get cancelling search item text.
    else if (lstrcmpi(pszParam, TEXT("/endsearch")) == 0)
    {
      g_pszParams[3] = (TCHAR*)GlobalAlloc(GPTR, sizeof(TCHAR)*string_size);
      popstring(g_pszParams[3]);
    }
    // Get cancelling process monitoring item text.
    else if (lstrcmpi(pszParam, TEXT("/endmonitor")) == 0)
    {
      g_pszParams[4] = (TCHAR*)GlobalAlloc(GPTR, sizeof(TCHAR)*string_size);
      popstring(g_pszParams[4]);
    }
    // Get caption.
    else if (lstrcmpi(pszParam, TEXT("/caption")) == 0)
    {
      popstring(pszParam);
      SetWindowText(g_hWndParent, pszParam);
    }
    // Allow next button?
    else if (lstrcmpi(pszParam, TEXT("/ignore")) == 0)
    {
      g_pszParams[5] = (TCHAR*)GlobalAlloc(GPTR, sizeof(TCHAR)*string_size);
      popstring(g_pszParams[5]);
    }
    // Auto close/kill on next click?
    else if (lstrcmpi(pszParam, TEXT("/autoclose")) == 0 || lstrcmpi(pszParam, TEXT("/autoclosesilent")) == 0)
    {
      g_bAutoClose = (lstrcmpi(pszParam, TEXT("/autoclose")) == 0 ? TRUE : 2);
      if (g_bAutoClose == TRUE)
      {
        g_pszParams[7] = (TCHAR*)GlobalAlloc(GPTR, sizeof(TCHAR)*string_size);
        popstring(g_pszParams[7]);
        g_pszParams[8] = (TCHAR*)GlobalAlloc(GPTR, sizeof(TCHAR)*string_size);
        popstring(g_pszParams[8]);
      }
      g_pszParams[6] = (TCHAR*)GlobalAlloc(GPTR, sizeof(TCHAR)*string_size);
      popstring(g_pszParams[6]);
      if (g_pszParams[5] == NULL)
      {
        g_pszParams[5] = (TCHAR*)GlobalAlloc(GPTR, sizeof(TCHAR)*string_size);
        ZeroMemory(g_pszParams[5], sizeof(TCHAR)*string_size);
      }
    }
    // Column headings text?
    else if (lstrcmpi(pszParam, TEXT("/colheadings")) == 0)
    {
      g_pszParams[9] = (TCHAR*)GlobalAlloc(GPTR, sizeof(TCHAR)*string_size);
      popstring(g_pszParams[9]);
      g_pszParams[10] = (TCHAR*)GlobalAlloc(GPTR, sizeof(TCHAR)*string_size);
      popstring(g_pszParams[10]);
    }
    // Use user icons rather than shell icons.
    else if (lstrcmpi(pszParam, TEXT("/usericons")) == 0)
      g_bUserIcons = TRUE;
    // End of parameters.
    else
    {
      pushstring(pszParam);
      break;
    }
  }
  GlobalFree(pszParam);

  if (!EnableDebugPriv())
  {
    pushstring(OUT_ERROR);
    return;
  }

  g_hDialog = CreateDialog(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG), g_hWndParent, (DLGPROC)DlgProc);
}

// Displays our dialog.
void ShowDialog()
{
  RECT dialog_r, r;
  LVCOLUMN lvc;
  LVITEM lvi;
  HICON hIconInfo, hIconSearch, hIconDefault;
  HIMAGELIST hImageList;

  // Set dialog font to that of the parent window font.
  SendMessage(g_hDialog, WM_SETFONT, (WPARAM)SendMessage(g_hWndParent, WM_GETFONT, 0, 0), TRUE);

  // Get the sizes of the UI.
  GetWindowRect(GetDlgItem(g_hWndParent, IDC_BLACKRECT), &dialog_r);
  MapWindowPoints(HWND_DESKTOP, g_hWndParent, (LPPOINT)&dialog_r, 2);

  // Set our window size to fit the UI size.
  MoveWindow(
    g_hDialog,
    dialog_r.left,
    dialog_r.top,
    dialog_r.right - dialog_r.left,
    dialog_r.bottom - dialog_r.top,
    FALSE
  );

  g_hProgress = GetDlgItem(g_hDialog, IDC_PROGRESS);
  g_hList = GetDlgItem(g_hDialog, IDC_LIST);
  g_hNext = GetDlgItem(g_hWndParent, 1);
  g_hCancel = GetDlgItem(g_hWndParent, 2);
  g_hBack = GetDlgItem(g_hWndParent, 3);

  // Disable next button while searching.
  EnableWindow(g_hNext, FALSE);

  // Resize controls.
  HWND hHeading = GetDlgItem(g_hDialog, IDC_HEADING);
  GetWindowRect(hHeading, &r);
  SetWindowPos(hHeading, 0, 0, 0, dialog_r.right - dialog_r.left, r.bottom - r.top, SWP_NOMOVE | SWP_NOZORDER);
  GetWindowRect(g_hProgress, &r);
  SetWindowPos(g_hProgress, 0, 0, 0, dialog_r.right - dialog_r.left, r.bottom - r.top, SWP_NOMOVE | SWP_NOZORDER);
  GetWindowRect(g_hList, &r);
  MapWindowPoints(HWND_DESKTOP, g_hWndParent, (LPPOINT)&r, 2);
  SetWindowPos(g_hList, 0, 0, 0, dialog_r.right - dialog_r.left, dialog_r.bottom - r.top, SWP_NOMOVE | SWP_NOZORDER);

  // Add columns.
  GetClientRect(g_hList, &r);
  lvc.mask = LVCF_TEXT | LVCF_WIDTH;
  lvc.cx = int((r.right - GetSystemMetrics(SM_CXHSCROLL))/3*2)-1;
  lvc.pszText = (g_pszParams[9] && g_pszParams[9][0] ? g_pszParams[9] : TEXT_COLHEADING_1);
  ListView_InsertColumn(g_hList, 0, &lvc);
  lvc.cx = int((r.right - GetSystemMetrics(SM_CXHSCROLL))/3*1)-1;
  lvc.pszText = (g_pszParams[10] && g_pszParams[10][0] ? g_pszParams[10] : TEXT_COLHEADING_2);
  ListView_InsertColumn(g_hList, 1, &lvc);

  // Set heading text.
  SetWindowText(GetDlgItem(g_hDialog, IDC_HEADING), g_pszParams[0] ? g_pszParams[0] : TEXT_HEADING);

  // Full row select always.
  ListView_SetExtendedListViewStyle(g_hList, LVS_EX_FULLROWSELECT | LVS_EX_LABELTIP);

  // Get info icon.
  if (g_bUserIcons)
    hIconInfo = (HICON)LoadImage(NULL, ICON_INFO, IMAGE_ICON, 0, 0, LR_DEFAULTSIZE|LR_LOADFROMFILE);
  else
    hIconInfo = ExtractIcon(g_hInstance, "shell32.dll", 15);
  if (!hIconInfo)
    hIconInfo = (HICON)LoadImage(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_INSTALLERICON), IMAGE_ICON, 0, 0, LR_DEFAULTSIZE);

  // Get searching icon.
  if (g_bUserIcons)
    hIconSearch = (HICON)LoadImage(NULL, ICON_SEARCH, IMAGE_ICON, 0, 0, LR_DEFAULTSIZE|LR_LOADFROMFILE);
  else
    hIconSearch = ExtractIcon(g_hInstance, "shell32.dll", 22);
  if (!hIconSearch)
    hIconSearch = hIconInfo;

  // Get default app icon.
  hIconDefault = ExtractIcon(g_hInstance, "shell32.dll", 2);

  // Create the icons image list.
  hImageList = ImageList_Create(16, 16, ILC_MASK|ILC_COLOR32, 8, 56);

  // Add icons to the image list.
  ImageList_AddIcon(hImageList, hIconInfo);
  ImageList_AddIcon(hImageList, hIconSearch);
  ImageList_AddIcon(hImageList, hIconDefault);
  DestroyIcon(hIconInfo);
  if (hIconSearch != hIconInfo)
    DestroyIcon(hIconSearch);
  DestroyIcon(hIconDefault);

  // Set the image list.
  ListView_SetImageList(g_hList, hImageList, LVSIL_SMALL);

  // Add 'Searching, please wait...'
  lvi.mask = LVIF_PARAM | LVIF_TEXT | LVIF_IMAGE;
  lvi.iItem = 0;
  lvi.iSubItem = 0;
  lvi.pszText = (g_pszParams[2] ? g_pszParams[2] : TEXT_SEARCHING_0);
  lvi.lParam = NULL;
  lvi.iImage = 1;
  ListView_InsertItem(g_hList, &lvi);
  ListView_EnsureVisible(g_hList, 0, FALSE);

  ParentDlgProcOld = (DLGPROC)SetWindowLongPtr(g_hWndParent, DWLP_DLGPROC, (LONG)ParentDlgProc);
  ListWndProcOld = (WNDPROC)SetWindowLongPtr(g_hList, GWLP_WNDPROC, (LONG)ListWndProc);

  // Tell NSIS to remove old inner dialog and pass handle of the new inner dialog.
  SendMessage(g_hWndParent, WM_NOTIFY_CUSTOM_READY, (WPARAM)g_hDialog, 0);
  ShowWindow(g_hDialog, SW_SHOW);

  g_done = FALSE;

  // Begin the thread of doom.
  g_hThreadFiles = CreateThread(NULL, 0, FilesThread, NULL, 0, NULL);

  // Loop until the user clicks on a button.
  while (!g_done || WaitForSingleObject(g_hThreadFiles, 0) == WAIT_TIMEOUT)
  {
    MSG msg;
    int nResult = GetMessage(&msg, NULL, 0, 0);
    if (!IsDialogMessage(g_hDialog, &msg) && !IsDialogMessage(g_hWndParent, &msg))
    {
      TranslateMessage(&msg);
      DispatchMessage(&msg);
    }
  }

  CloseHandle(g_hThreadFiles);

  // Free allocated memory.
  FreeGlobalAllocs();

  // Set window dialog procedure back to NSIS's.
  SetWindowLongPtr(g_hWndParent, DWLP_DLGPROC, (LONG)ParentDlgProcOld);
  DestroyWindow(g_hDialog);

  // Return page button result.
  pushstring(g_is_cancel ? OUT_CANCEL : g_is_back ? OUT_BACK : OUT_NEXT);
}

// NSIS Function: Create dialog but do not show it.
extern "C"
void __declspec(dllexport) InitDialog(HWND hWndParent, int string_size, char *variables, stack_t **stacktop, extra_parameters *extra)
{
  g_hWndParent = hWndParent;
  EXDLL_INIT();
  {
    if (!g_hDialog)
    {
      LoadDialog(string_size);

      // Return page HWND.
      TCHAR szHWND[32];
      wsprintf(szHWND, "%d", g_hDialog);
      pushstring(szHWND);
    }
    else
      pushstring(OUT_ERROR);
  }
}

// NSIS Function: Add a file.
extern "C"
void __declspec(dllexport) AddFile(HWND hWndParent, int string_size, char *variables, stack_t **stacktop, extra_parameters *extra)
{
  g_hWndParent = hWndParent;
  EXDLL_INIT();
  {
    if (FilesCount < MAX_LIST_COUNT)
    {
      FilesList[FilesCount] = (TCHAR*)GlobalAlloc(GPTR, sizeof(TCHAR)*MAX_PATH);
      popstring(FilesList[FilesCount]);
      CharLower(FilesList[FilesCount]);
      FilesCount++;
    }
    else
      pushstring(OUT_ERROR);
  }
}

// NSIS Function: Add a module.
extern "C"
void __declspec(dllexport) AddModule(HWND hWndParent, int string_size, char *variables, stack_t **stacktop, extra_parameters *extra)
{
  g_hWndParent = hWndParent;
  EXDLL_INIT();
  {
    if (ModulesCount < MAX_LIST_COUNT)
    {
      ModulesList[ModulesCount] = (TCHAR*)GlobalAlloc(GPTR, sizeof(TCHAR)*MAX_PATH);
      popstring(ModulesList[ModulesCount]);
      CharLower(ModulesList[ModulesCount]);
      ModulesCount++;
    }
    else
      pushstring(OUT_ERROR);
  }
}

// NSIS Function: Add a window class.
extern "C"
void __declspec(dllexport) AddClass(HWND hWndParent, int string_size, char *variables, stack_t **stacktop, extra_parameters *extra)
{
  g_hWndParent = hWndParent;
  EXDLL_INIT();
  {
    if (ClassesCount < MAX_LIST_COUNT)
    {
      ClassesList[ClassesCount] = (TCHAR*)GlobalAlloc(GPTR, sizeof(TCHAR)*MAX_PATH);
      popstring(ClassesList[ClassesCount]);
      CharLower(ClassesList[ClassesCount]);
      ClassesCount++;
    }
    else
      pushstring(OUT_ERROR);
  }
}

// NSIS Function: Add a window caption.
extern "C"
void __declspec(dllexport) AddCaption(HWND hWndParent, int string_size, char *variables, stack_t **stacktop, extra_parameters *extra)
{
  g_hWndParent = hWndParent;
  EXDLL_INIT();
  {
    if (CaptionsCount < MAX_LIST_COUNT)
    {
      CaptionsList[CaptionsCount] = (TCHAR*)GlobalAlloc(GPTR, sizeof(TCHAR)*MAX_PATH);
      popstring(CaptionsList[CaptionsCount]);
      CharLower(CaptionsList[CaptionsCount]);
      CaptionsCount++;
    }
    else
      pushstring(OUT_ERROR);
  }
}

// NSIS Function: Show all open programs.
extern "C"
void __declspec(dllexport) AddApplications(HWND hWndParent, int string_size, char *variables, stack_t **stacktop, extra_parameters *extra)
{
  g_hWndParent = hWndParent;
  EXDLL_INIT();
  {
    g_bShowRunningApplications = TRUE;
  }
}

// NSIS Function: Display the initialised dialog.
extern "C"
void __declspec(dllexport) Show(HWND hWndParent, int string_size, char *variables, stack_t **stacktop, extra_parameters *extra)
{
  g_hWndParent = hWndParent;
  EXDLL_INIT();
  {
    if (g_hDialog)
      ShowDialog();
    else
      pushstring(OUT_ERROR);
  }
}

// NSIS Function: Display the dialog.
extern "C"
void __declspec(dllexport) Dialog(HWND hWndParent, int string_size, char *variables, stack_t **stacktop, extra_parameters *extra)
{
  g_hWndParent = hWndParent;
  EXDLL_INIT();
  {
    if (!g_hDialog)
    {
      LoadDialog(string_size);
      ShowDialog();
    }
    else
      pushstring(OUT_ERROR);
  }
}

// NSIS Function: Search without a dialog.
extern "C"
void __declspec(dllexport) SilentSearch(HWND hWndParent, int string_size, char *variables, stack_t **stacktop, extra_parameters *extra)
{
  g_hWndParent = hWndParent;
  EXDLL_INIT();
  {
    TCHAR *pszParam = (TCHAR*)LocalAlloc(LPTR, sizeof(TCHAR)*string_size);

    if (!EnableDebugPriv())
    {
      pushstring(OUT_ERROR);
      return;
    }

    if (popstring(pszParam) == 0)
    {
      // Perform the search with a separate thread.
      if (lstrcmpi(pszParam, "/thread") == 0)
        g_hThreadFiles = CreateThread(NULL, 0, FilesThreadSilent, NULL, 0, NULL);
      // Perform the search without a separate thread.
      else
      {
        pushstring(pszParam);
        FilesThreadSilent(0);
        PushStrings();
      }
    }
    // Perform the search without a separate thread.
    else
    {
      FilesThreadSilent(0);
      PushStrings();
    }

    LocalFree(pszParam);
  }
}

// NSIS Function: Wait for thread to finish.
extern "C"
void __declspec(dllexport) SilentWait(HWND hWndParent, int string_size, char *variables, stack_t **stacktop, extra_parameters *extra)
{
  g_hWndParent = hWndParent;
  EXDLL_INIT();
  {
    if (g_hThreadFiles)
    {
      TCHAR *pszParam = (TCHAR*)LocalAlloc(LPTR, sizeof(TCHAR)*string_size);

      if (popstring(pszParam) == 0)
      {
        // Wait for so many milliseconds.
        if (lstrcmpi(pszParam, "/time") == 0)
        {
          popstring(pszParam);
          int ms = my_atoi(pszParam);
          if (WaitForSingleObject(g_hThreadFiles, ms) == WAIT_TIMEOUT)
            pushstring(OUT_WAIT);
          else
          {
            PushStrings();
            pushstring(OUT_DONE);
          }
        }
        // Wait until the thread has finished.
        else
        {
          pushstring(pszParam);
          WaitForSingleObject(g_hThreadFiles, INFINITE);
          CloseHandle(g_hThreadFiles);
          PushStrings();
        }
      }
      // Wait until the thread has finished.
      else if (g_hThreadFiles)
      {
        WaitForSingleObject(g_hThreadFiles, INFINITE);
        CloseHandle(g_hThreadFiles);
        PushStrings();
      }

      LocalFree(pszParam);
    }
    else
      pushstring(OUT_ERROR);
  }
}

// NSIS Function: What is the % complete?
extern "C"
void __declspec(dllexport) SilentPercentComplete(HWND hWndParent, int string_size, char *variables, stack_t **stacktop, extra_parameters *extra)
{
  g_hWndParent = hWndParent;
  EXDLL_INIT();
  {
    TCHAR szPercentComplete[4];
    wsprintf(szPercentComplete, "%d", g_iPercentComplete);
    pushstring(szPercentComplete);
  }
}
extern "C"
void __declspec(dllexport) Unload(HWND hWndParent, int string_size, char *variables, stack_t **stacktop, extra_parameters *extra)
{
  EXDLL_INIT();
  {
    // Free allocated memory.
    FreeGlobalAllocs();
  }
}

// Entry point for DLL.
BOOL WINAPI DllMain(HINSTANCE hInst, ULONG ul_reason_for_call, LPVOID lpReserved)
{
  g_hInstance=hInst;
    return TRUE;
}

// Function: Converts char to int.
int my_atoi(TCHAR *p)
{
  int n=0, f=0;

  for (;;p++)
  {
    switch (*p)
    {
    case ' ':
    case '\t':
      continue;
    case '-':
      f++;
    case '+':
      p++;
    }
    break;
  }
  while(*p >= '0' && *p <= '9')
    n = n*10 + *p++ - '0';
  return(f? -n: n);
}

// Function: Search using wildcards.
int wildcmp(const TCHAR *wild, const TCHAR *string)
{
  // Written by Jack Handy - jakkhandy@hotmail.com
  // http://www.codeproject.com/KB/string/wildcmp.aspx

  const TCHAR *cp = NULL, *mp = NULL;

  while ((*string) && (*wild != '*'))
  {
    if ((*wild != *string) && (*wild != '?'))
      return 0;
    wild++;
    string++;
  }

  while (*string)
  {
    if (*wild == '*')
    {
      if (!*++wild)
        return 1;
      mp = wild;
      cp = string+1;
    }
    else if ((*wild == *string) || (*wild == '?'))
    {
      wild++;
      string++;
    }
    else
    {
      wild = mp;
      string = cp++;
    }
  }

  while (*wild == '*')
    wild++;
  return !*wild;
}
